#!/usr/bin/env python3

import argparse
import sys
import time
from collections import defaultdict
from multiprocessing import Process, Queue
from index_registry.dynamodb.stats_table import StatsTable
#from index_registry.dynamodb.lease_table import LeaseTable
from index_registry.utils.config import Config

QUEUE_SENTINEL_OBJECT = "sentinel"


def timeit(func):
    def timed(*args, **kw):
        start = time.time()
        res = func(*args, **kw)
        print(f"{func.__name__} took {time.time() - start} seconds")
        return res
    return timed


def get_table_pk_values_segment(table, pk, result_queue, segment, total_segments):
    response = table.scan(
        Segment=segment,
        TotalSegments=total_segments,
        ProjectionExpression=pk
    )

    result_queue.put_nowait([StatsTable.get_dynamo_fields(resp[pk]) for resp in response["Items"]])


@timeit
def get_table_pk_values(table, pk, result_queue, process_count):
    process_pool = [Process(target=get_table_pk_values_segment, args=(table, pk, result_queue, i, process_count)) for i in range(process_count)]

    # Start the threads
    for process in process_pool:
        process.start()

    # Block main thread until all threads are finished
    for process in process_pool:
        process.join()

    result_queue.put_nowait(QUEUE_SENTINEL_OBJECT)


def truncate_table_segment(table, pk, segment, total_segments):
    response = table.scan(
        Segment=segment,
        TotalSegments=total_segments,
        ProjectionExpression=pk
    )

    assert response.get('LastEvaluatedKey', None) is None

    # TODO: What happens to the Segments after we start deleting items? If we have 10 Segments in the beginning, then delete X items, will this reorder the segments?
    # TODO: This batch writer is used by all processes. Try a table per process.
    with table.batch_writer() as batch:
        for index in response['Items']:
            batch.delete_item(
                Key={
                    pk: index[pk],
                }
            )


@timeit
def truncate_table(table, pk, process_count):
    process_pool = [Process(target=truncate_table_segment, args=(table, pk, i, process_count)) for i in range(process_count)]

    # Start threads
    for process in process_pool:
        process.start()

    # Block main thread until all threads are finished
    for process in process_pool:
        process.join()

    # Verify the table is truncated
    assert len(table.scan(
        ProjectionExpression=pk,
        Limit=1
    )["Items"]) == 0


@timeit
def aggregate_raw_data(entry_queue, output_queue):
    orgs_stats = defaultdict(lambda: defaultdict(int))
    lease_indexes_counter = 0

    while True:
        dynamo_fields_list = entry_queue.get()
        if dynamo_fields_list == QUEUE_SENTINEL_OBJECT:
            break

        lease_indexes_counter += len(dynamo_fields_list)
        for dynamo_fields in dynamo_fields_list:
            org_id, count_metric, size_metric, index_size = dynamo_fields

            orgs_stats[org_id][count_metric] += 1
            orgs_stats[org_id][size_metric] += index_size

    for org in orgs_stats.items():
        output_queue.put_nowait(org)

    output_queue.put_nowait(QUEUE_SENTINEL_OBJECT)

    print(f'{lease_indexes_counter} records were scanned from the lease table')
    print(f'The sync will be successful if the stats table has {len(orgs_stats)} records when finished')


def write_stats_segment(stats_table, org_queue):
    with stats_table.table.batch_writer() as batch:
        while True:
            res = org_queue.get()
            if res == QUEUE_SENTINEL_OBJECT:
                org_queue.put_nowait(QUEUE_SENTINEL_OBJECT)
                break

            org_id, org_data = res
            org_data["org_id"] = org_id
            batch.put_item(org_data)


@timeit
def write_stats(stats_table, org_queue, process_count):
    process_pool = [Process(target=write_stats_segment, args=(stats_table, org_queue)) for _ in range(process_count)]

    # Start threads
    for process in process_pool:
        process.start()

    # Block main thread until all threads are finished
    for process in process_pool:
        process.join()


def _get_config(args):
    config = Config.create()

    config.set('aws', 'default_region', args.aws_region)
    config.set('lease_dynamodb', 'dynamodb_table', args.lease_dynamodb_table)
    config.set('stats', 'dynamodb_table', args.stats_dynamodb_table)

    return config


def build_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument('--aws-region', dest='aws_region',
                        type=str, required=True, default='us-east-1',
                        help='AWS Region')
    parser.add_argument('--lease-table', dest='lease_dynamodb_table',
                        type=str, required=True, default='dev__index_leases',
                        help='The lease dynamodb table name')
    parser.add_argument('--stats-table', dest='stats_dynamodb_table',
                        type=str, required=True, default='dev__index_stats',
                        help='Keep track of the index statistics')
    parser.add_argument('--stats-processes', dest='stats_processes',
                        type=int, required=True, default=60,
                        help='The number of processes that will be deleting the stats table items')
    parser.add_argument('--lease-processes', dest='lease_processes',
                        type=int, required=True, default=60,
                        help='The number of processes that will be working with the lease table.')

    return parser


@timeit
def sync(config, lease_processes, stats_processes):
    stats_table = StatsTable(config)
    #lease_table = LeaseTable(None, config)

    raw_data = Queue()
    aggregated_data = Queue()

    # Delete the stats table
    trunc_proc = Process(target=truncate_table, args=(stats_table.table, "org_id", stats_processes))
    trunc_proc.start()

    # Scan the lease table and put all raw records in the raw_data queue

    # TODO: Comment out for now, this will need to get re-worked for mongo, and for sure do not like how this directly access the table.
    # read_proc = Process(target=get_table_pk_values, args=(lease_table.table, "index_path", raw_data, lease_processes))
    # read_proc.start()

    # Process the records from the lease table and create statistics for each org
    agg_proc = Process(target=aggregate_raw_data, args=(raw_data, aggregated_data))
    agg_proc.start()

    # Make sure the stats table is now empty
    trunc_proc.join()

    # Write the statistics in the stats table
    write_stats(stats_table, aggregated_data, stats_processes)

    #read_proc.join()
    agg_proc.join()


@timeit
def main():
    # Set up the config from the flags
    parser = build_parser()
    args = parser.parse_args()
    config = _get_config(args)

    sync(config, args.lease_processes, args.stats_processes)


if __name__ == "__main__":
    sys.exit(main())
