import time
import logging
import pymongo
from pymongo import MongoClient
from index_registry.utils.index_field_helpers import get_size_from_props, properties_from_key
# from pymongo.write_concern import WriteConcern

_logger = logging.getLogger(__name__)


class DbManager:
    def __init__(self, config):
        host = config.get("mongo", "host")
        port = config.get("mongo", "port")
        user = config.get("mongo", "user")
        password = config.get("mongo", "password")
        db_name = config.get("mongo", "db_name")
        collection_name = config.get("lease_mongodb", "mongodb_collection")
        lease_cache_size = config.getint("lease_mongodb", "lease_cache_size")
        self.lease_extension_seconds = config.getint("lease_mongodb", "lease_extension_seconds")
        direct_connect = config.getboolean("mongo", "direct_connect")
        local_dev_db = config.getboolean("mongo", "local_dev_db")
        if user == "none":
            uri = f"mongodb://{host}:{port}"
        elif direct_connect:
            uri = f"mongodb://{user}:{password}@{host}:{port}/{db_name}?directConnection=true"
        else:
            uri = f"mongodb://{user}:{password}@{host}:{port}/{db_name}"
        client = MongoClient(uri)
        self.db = client[db_name]
        self.db_admin = client.admin
        self.ack_client = self.db[collection_name]
        # If running locally attempt to initialize the DB, this happens with a separate job in prod.
        if local_dev_db:
            self.init_db()
        # Unacked client is not used currently but leaving for now in case we change our minds
        # self.unacked_client  = self.db.leases.with_options(write_concern=WriteConcern(w=0))

    def init_db(self):
        self.ack_client.create_index([("index_path", pymongo.ASCENDING)], unique=True)
        self.ack_client.create_index([("is_merge", pymongo.ASCENDING)], unique=False)
        self.ack_client.create_index([("lease_expiration", pymongo.ASCENDING)], unique=False)
        self.ack_client.create_index([("container_id", pymongo.ASCENDING)], unique=False)
        self.ack_client.create_index([("index_size", pymongo.ASCENDING)], unique=False)

    def delete_db(self):
        # Added for testing ease
        self.ack_client.drop()

    def add_index(self, index_path, is_merge):
        size = get_size_from_props(properties_from_key(index_path))
        post = {"index_path": index_path, "is_merge": is_merge, "index_size": size}
        self.ack_client.insert_one(post)
        _logger.info(f"Added index {index_path}")

    def remove_index(self, index_path, is_merge):
        # Not using is_merge, index_path is unique is_merge is just included to future-proof.
        post = {"index_path": index_path}
        self.ack_client.delete_one(post)
        _logger.info(f"Deleted index {index_path} is_merge {is_merge}")

    def get_index(self, index_path):
        # Not expected to be used, but adding for testing ease
        query = {"index_path": index_path}
        projection = {"_id": 0}
        return self.ack_client.find_one(query, projection)

    def entry_exists(self, index_path):
        if self.get_index(index_path):
            return True
        return False

    def extend_leases(self, container_id):
        new_lease_expiration = round(time.time()) + self.lease_extension_seconds
        post = {"lease_expiration": new_lease_expiration}
        ret = self.ack_client.update_many({"container_id": container_id}, {"$set": post}, upsert=False)
        _logger.info(f"Extended {ret.modified_count} leases for {container_id}")

    def get_existing_leases(self, container_id):
        query = {"$and": [{"is_merge": {"$eq": True}}, {"container_id": {"$eq": container_id}}]}
        merger_index_paths = self._query_index_paths(query)
        query = {"$and": [{"is_merge": {"$eq": False}}, {"container_id": {"$eq": container_id}}]}
        optimized_index_paths = self._query_index_paths(query)
        return merger_index_paths, optimized_index_paths

    def get_new_lease(self, container_id, is_merge, mb_limit):
        current_time = round(time.time())
        lease_expiration = current_time + self.lease_extension_seconds
        post = {"lease_expiration": lease_expiration, "container_id": container_id, "hosted": False}
        query = {"$and": [{"is_merge": is_merge},
                          {"index_size": {"$lt": mb_limit}},
                          {"$or": [{"lease_expiration": {"$lt": current_time}},
                                   {"lease_expiration": {"$exists": False}}]}]}
        projection = {"index_path": 1, "_id": 0}
        ret = self.ack_client.find_one_and_update(query, {"$set": post}, upsert=False, projection=projection)
        if ret:
            return ret['index_path']
        return ""

    def release_leases(self, container_id):
        post = {"lease_expiration": 0,
                "container_id": "none",
                "hosted": False}
        ret = self.ack_client.update_many({"container_id": container_id},
                                          {"$set": post}, upsert=False)
        _logger.info(f"Released {ret.modified_count} leases for {container_id}")

    def release_lease(self, container_id, is_merge, index_path):
        # Ignoring is_merge for now update query is the same either way since index_path is unique.
        post = {"lease_expiration": 0,
                "container_id": "none",
                "hosted": False}
        query = {"$and": [{"container_id": container_id}, {"index_path": index_path}]}
        ret = self.ack_client.update_one(query, {"$set": post}, upsert=False)
        if ret.modified_count != 1:
            _logger.warning(f"Attempted to unlease Index Path {index_path} not currently leased by {container_id}")
            return False
        _logger.info(f"Release single lease for {container_id} path {index_path} is_merge {is_merge}")
        return True

    def iterate_all_index_paths(self):
        projection = {"index_path": 1, "_id": 0}
        for ret in self.ack_client.find(projection=projection):
            yield ret['index_path']

    def confirm_hosted(self, container_id, is_merge, index_path):
        # Ignoring is_merge for now update query is the same either way since index_path is unique.
        query = {"$and": [{"container_id": container_id}, {"index_path": index_path}]}
        post = {"hosted": True}
        ret = self.ack_client.update_one(query, {"$set": post}, upsert=False)
        if ret.modified_count != 1:
            _logger.warning(f"Index path {index_path} did not update to hosted"
                            f"Possible lease has been stolen or some indexes were already flagged hosted")
            return False
        _logger.info(f"Confirming hosted for {container_id}, index_path {index_path} and is_merge {is_merge}")
        return True

    def check_hosted(self, container_id, merging_index_paths, optimized_index_paths):
        # container_id is passed for completeness but not used in this method
        or_list = self._build_or_list(merging_index_paths, optimized_index_paths)
        query = {"$or": or_list}
        projection = {"index_path": 1, "hosted": 1, "is_merge": 1, "container_id": 1, "_id": 0}
        ret = self.ack_client.find(query, projection)
        merging_index_path_statuses = {}
        optimized_index_path_statuses = {}
        for row in ret:
            if row["is_merge"]:
                merging_index_path_statuses[row["index_path"]] = {
                    "container_id": row["container_id"] if "container_id" in row else None,
                    "hosted": row["hosted"] if "hosted" in row else None}
            else:
                optimized_index_path_statuses[row["index_path"]] = {
                    "container_id": row["container_id"] if "container_id" in row else None,
                    "hosted": row["hosted"] if "hosted" in row else None}
        _logger.info(f"Returning {len(merging_index_path_statuses)} merging hosted and "
                     f"{len(optimized_index_path_statuses)} optimized hosted to {container_id}")
        return merging_index_path_statuses, optimized_index_path_statuses

    @staticmethod
    def _build_or_list(merging_index_paths, optimized_index_paths):
        or_list = []
        for merging_index_path in merging_index_paths:
            or_list.append({"index_path": merging_index_path})
        for optimized_index_path in optimized_index_paths:
            or_list.append({"index_path": optimized_index_path})
        return or_list

    def _query_index_paths(self, query, limit=None):
        projection = {"index_path": 1, "_id": 0}
        index_paths = []
        if limit:
            ret = self.ack_client.find(query, projection).limit(limit)
        else:
            ret = self.ack_client.find(query, projection)
        for row in ret:
            index_paths.append(row["index_path"])
        return index_paths

    def _query_counts(self, query):
        return self.ack_client.count_documents(query)

    def _get_count(self, is_merge):
        query = {"is_merge": is_merge}
        return self._query_counts(query)

    def get_merging_count(self):
        return self._get_count(True)

    def get_optimized_count(self):
        return self._get_count(False)

    def _get_leased(self, is_merge):
        current_time = time.time()
        query = {"$and": [{"is_merge": {"$eq": is_merge}},
                          {"lease_expiration": {"$gt": current_time}}]}
        return self._query_counts(query)

    def get_merging_leased(self):
        return self._get_leased(True)

    def get_optimized_leased(self):
        return self._get_leased(False)

    def _get_hosted(self, is_merge):
        current_time = time.time()
        query = {"$and": [{"is_merge": {"$eq": is_merge}},
                          {"lease_expiration": {"$gt": current_time}},
                          {"hosted": {"$eq": True}}]}
        return self._query_counts(query)

    def get_merging_hosted(self):
        return self._get_hosted(True)

    def get_optimized_hosted(self):
        return self._get_hosted(False)
