import logging
from fastapi import FastAPI, HTTPException, Response, status
from typing import Dict, Any
from pydantic import BaseModel
from typing import Optional

from pymongo.errors import DuplicateKeyError

from index_registry.engine.index_engine import IndexEngine
from index_registry.engine.searcher_engine import SearcherEngine, SearcherRequest
from index_registry.engine.org_sqs_engine import OrgSqsEngine, IndexLeaseEntryDoesNotExist
from index_registry.mongodb.orgsqs_query_entry import QueryEntry, QueryEntries
from index_registry.mongodb.orgsqs_db_interface import IndexTimestamps
from index_registry.mongodb.leases_db_interface import DbManager as LeasesDbManager
from index_registry.utils.config import Config

from pscr_common.utils import datetime_from_ms_timestamp

_logger = logging.getLogger(__name__)


class TooManyBulkItems(HTTPException):
    pass


def build_routes(app: FastAPI, config: Config,
                 leases_db_manager: LeasesDbManager,
                 index_engine: IndexEngine,
                 searcher_engine: SearcherEngine,
                 org_sqs_engine: OrgSqsEngine):
    """Index Registry API routes and their handlers"""

    @app.get("/internal/registry/healthcheck")
    def healthcheck():
        """Dummy healthcheck"""
        return {"status": 200}

    class GetNewLease(BaseModel):
        is_merge: bool
        mb_limit: int

    @app.post("/internal/registry/searcher/v1/get_new_lease/{container_id}")
    def leases_get_new_lease(container_id: str, user_data: GetNewLease):
        _logger.info(f"Getting new lease for {container_id}")
        lease = leases_db_manager.get_new_lease(container_id, user_data.is_merge, user_data.mb_limit)
        _logger.info(f"Returning lease for {lease if lease != '' else 'NO_LEASE_AVAILABLE'} to {container_id}")
        return {"index_path": lease}

    @app.post("/internal/registry/searcher/v1/extend_leases/{container_id}")
    def leases_extend_leases(container_id: str):
        _logger.info(f"Extending Leases for {container_id}")
        leases_db_manager.extend_leases(container_id)

    @app.post("/internal/registry/searcher/v1/get_existing_leases/{container_id}")
    def leases_get_existing_leases(container_id: str):
        _logger.info(f"Getting existing Leases for {container_id}")
        merging_indexes, optimized_indexes = leases_db_manager.get_existing_leases(container_id)
        _logger.info(f"Leases for {container_id}, It has {len(merging_indexes)} merging indexes and "
                     f"{len(optimized_indexes)} optimized indexes")
        return {"merger_index_paths": merging_indexes, "optimized_index_paths": optimized_indexes}

    @app.post("/internal/registry/searcher/v1/release_leases/{container_id}")
    def leases_release_leases(container_id: str):
        _logger.info(f"Releasing all leases for {container_id}")
        leases_db_manager.release_leases(container_id)

    class IndexPath(BaseModel):
        is_merge: bool
        index_path: str

    @app.post("/internal/registry/searcher/v1/release_lease/{container_id}")
    def leases_release_lease(container_id: str, user_data: IndexPath):
        _logger.info(f"Releasing lease for {container_id}, path {user_data.index_path}")
        leases_db_manager.release_lease(container_id, user_data.is_merge, user_data.index_path)

    @app.post("/internal/registry/searcher/v1/confirm_hosted/{container_id}")
    def leases_confirm_hosted(container_id: str, user_data: IndexPath):
        _logger.info(f"Confirming hosted lease for {user_data.index_path} on {container_id}")
        leases_db_manager.confirm_hosted(container_id, user_data.is_merge, user_data.index_path)

    class CheckHosted(BaseModel):
        merging_index_paths: tuple
        optimized_index_paths: tuple

    @app.post("/internal/registry/searcher/v1/check_hosted/{container_id}")
    def leases_check_hosted(container_id: str, user_data: CheckHosted):
        _logger.info(f"{container_id} is checking to see if "
                     f"{len(user_data.merging_index_paths)} merging indexes is hosted and "
                     f"{len(user_data.optimized_index_paths)} optimized indexes is hosted")
        merge_statuses, optimized_statuses = leases_db_manager.check_hosted(container_id,
                                                                            user_data.merging_index_paths,
                                                                            user_data.optimized_index_paths)
        return {"merging_index_path_statuses": merge_statuses,
                "optimized_index_path_statuses": optimized_statuses}

    @app.post("/internal/registry/searcher/v1/add_index_lease", status_code=201)
    def leases_add_lease(user_data: IndexPath, response: Response):
        _logger.info(f"Adding index {user_data.index_path}")
        try:
            leases_db_manager.add_index(user_data.index_path, user_data.is_merge)
        except DuplicateKeyError as ex:
            _logger.warning("Caught duplicate key error, returning successful insert: %s", ex)
            response.status_code = status.HTTP_200_OK

    @app.post("/internal/registry/searcher/v1/remove_index_lease")
    def leases_delete_lease(user_data: IndexPath):
        _logger.info(f"Deleting index {user_data.index_path}")
        leases_db_manager.remove_index(user_data.index_path, user_data.is_merge)


    # @app.post("/internal/registry/index/v1/sync")
    # def index_sync(params: Dict[Any, Any]):
    #     _logger.info("Index sync log called")
    #     org_id = params["org_id"]
    #     bucket = params["bucket"]
    #
    #     # returns {"added": NUM_INDEXES_ADDED, "removed": NUM_INDEXES_REMOVED}
    #     return index_engine.sync_from_s3(org_id, bucket)
    #
    # @app.post("/internal/registry/index/v1/expire")
    # def index_expire(params: Dict[Any, Any]):
    #     _logger.info("v1 expore called")
    #     index_path = params["index_path"]
    #
    #     return {"indexes": index_engine.delete_index(index_path)}


    # @app.post("/internal/registry/index/v1/expire/org")
    # def org_index_expire(params: Dict[Any, Any]):
    #     _logger.info("v1 org expire called")
    #     org_id = params["org_id"]
    #
    #     return {"indexes": index_engine.delete_org_indexes(org_id)}

    @app.post("/internal/registry/searcher/v1/register")
    def register_searcher(params: Dict[Any, Any]):
        _logger.info("registration called")
        _logger.debug(f"Register params {params}")
        registrations = params.get('registrations', None)
        if registrations is None:
            # If registrations list not sent assume single registration
            registrations = [params]

        searchers = []
        for registration in registrations:
            searcher_id = registration.get("searcher_id")
            org_id = str(registration.get("org_id"))
            container_version = registration.get("container_version")
            endpoint = registration.get("endpoint")

            start_time = registration.get("start_time")
            if start_time is not None:
                start_time = datetime_from_ms_timestamp(start_time)

            end_time = registration.get("end_time")
            if end_time is not None:
                end_time = datetime_from_ms_timestamp(end_time)

            index_path = registration.get("index_path")

            last_search_time = registration.get('last_search_time')
            if last_search_time is not None:
                last_search_time = datetime_from_ms_timestamp(last_search_time)

            index_partition_id = registration.get("index_partition_id")
            index_stage = int(registration.get("index_stage")) if "index_stage" in registration else None

            searchers.append(SearcherRequest(searcher_id, org_id, container_version, endpoint,
                                            start_time, end_time, index_path, last_search_time,
                                            index_partition_id, index_stage))

        ret = {"searchers": searcher_engine.register_searchers(searchers)}
        # This is super ugly, but fast api does not like the SearcherResponse class returned further up
        # Not sure if this is worth fixing better if we plan to remove RDS anyway.
        for x in range(len(ret['searchers'])):
            ret['searchers'][x] = ret['searchers'][x].items()
        return ret

    @app.post("/internal/registry/searcher/v1/list")
    def list_searchers(params: Dict[Any, Any]):
        _logger.info("list searchers called")
        _logger.debug(f"List Searchers params {params}")
        org_id = params.get("org_id")
        start_time = datetime_from_ms_timestamp(params.get("start_time"))
        end_time = datetime_from_ms_timestamp(params.get("end_time"))
        index_path_filters = params.get("index_path_filters")
        index_partition_ids = params.get("index_partition_ids")

        ret = {"searchers": searcher_engine.get_searchers(str(org_id), start_time, end_time,
                                                          index_path_filters=index_path_filters,
                                                          index_partition_ids=index_partition_ids)}
        # This is super ugly, but fast api does not like the SearcherResponse class returned further up
        # Not sure if this is worth fixing better if we plan to remove RDS anyway.
        for x in range(len(ret['searchers'])):
            ret['searchers'][x] = ret['searchers'][x].items()
        return ret

    @app.post("/internal/registry/searcher/v1/limits")
    def get_limits(params: Dict[Any, Any]):
        _logger.info("limits called")
        _logger.debug(f"Limits params {params}")
        org_id = params.get("org_id")

        bounds = searcher_engine.get_searcher_time_bounds(str(org_id))

        return {"time_bounds": {"lower": bounds.lower, "upper": bounds.upper}}

    @app.post("/internal/registry/searcher/v1/validation_list")
    def get_validation_searchers(params: Dict[Any, Any]):
        _logger.info("validation_list called")
        _logger.debug(f"Validation List params {params}")
        org_id = params.get("org_id")
        start_time = datetime_from_ms_timestamp(params.get("start_time"))
        end_time = datetime_from_ms_timestamp(params.get("end_time"))

        ret = {"searchers": searcher_engine.get_validation_searchers(str(org_id), start_time, end_time)}
        # This is super ugly, but fast api does not like the SearcherResponse class returned further up
        # Not sure if this is worth fixing better if we plan to remove RDS anyway.
        for x in range(len(ret['searchers'])):
            ret['searchers'][x] = ret['searchers'][x].items()
        _logger.debug(f"registry.py/get_validation_searchers ret value: {ret}")
        return ret


    # used for internal validation only
    # should not be exposed to customers because we return stage 0 searchers from all orgs
    @app.post("/internal/registry/searcher/v1/validation_list/stage")
    def get_validation_searchers_by_stage(params: Dict[Any, Any]):
        _logger.info("validation_by_stage called")
        _logger.debug(f"Validation List by stage params {params}")
        stage = params.get("stage", 0)
        count = params.get("count", 3)

        ret = {"searchers": searcher_engine.get_validation_searchers_by_stage(stage, count)}
        # This is super ugly, but fast api does not like the SearcherResponse class returned further up
        # Not sure if this is worth fixing better if we plan to remove RDS anyway.
        for x in range(len(ret['searchers'])):
            ret['searchers'][x] = ret['searchers'][x].items()
        _logger.debug(f"registry.py/get_validation_searchers ret value: {ret}")
        return ret

    @app.post("/internal/registry/query/v1/add")
    def add_sqs_query(entry: QueryEntry):
        _logger.info(f"Adding orgsqs entry for {entry.consumer_name}")
        if entry.type == "raw":
            org_sqs_engine.org_sqs_add(entry)
        else:
            try:
                org_sqs_engine.org_sqs_add_if_valid_lease(entry)
            except IndexLeaseEntryDoesNotExist:
                return {"message": "Index Lease Entry Does not Exist did not insert entry"}
        return {}

    class QueryRetrieve(BaseModel):
        org_key: str
        index_timestamps: Optional[IndexTimestamps] = None

    @app.post("/internal/registry/query/v1/retrieve")
    def retrieve_sqs_query(user_data: QueryRetrieve):
        _logger.info(f"Orgsqs retrieve for {user_data.org_key} with {user_data.index_timestamps}")
        return org_sqs_engine.org_sqs_retrieve(user_data.org_key, user_data.index_timestamps)

    @app.post("/internal/registry/query/v1/retrieve_count")
    def retrieve_sqs_query_count(user_data: QueryRetrieve):
        _logger.info(f"Orgsqs retrieve count for {user_data.org_key} with {user_data.index_timestamps}")
        return {"entry_count": org_sqs_engine.org_sqs_retrieve_count(user_data.org_key, user_data.index_timestamps)}

    @app.post("/internal/registry/query/v1/add_bulk")
    def add_bulk_sqs(user_data: QueryEntries):
        _logger.info(f"Orgsqs Add Bulk called")
        _logger.debug(f"We have {len(user_data.root)} entries passed to bulk")
        max_bulk_items = config.getint("sqs", "max_bulk_adds", fallback=1000)
        if len(user_data.root) > max_bulk_items:
            raise TooManyBulkItems(status_code=500, detail=f"Too many Bulk items exceeds limit of {max_bulk_items}")
        org_sqs_engine.org_sqs_add_bulk(user_data)


